#include <err.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#if 500100 <= __FreeBSD_version
# include <sys/resource.h>
#endif
#if __FreeBSD_version <= 500100
# include <sys/dkstat.h>
#endif

#ifndef lint
static const char rcs_id[] = "$Id: speedstepd.c,v 1.6 2003/02/20 08:43:41 a-gota Exp $";
#endif


#define	DEFAULT_INTERVAL	  2
#define	MIN_INTERVAL		  1

#define	MAX_NUM_OF_DATA		 10
#define	DEFAULT_NUM_OF_DATA	  3
#define	MIN_NUM_OF_DATA		  1

#define	MAX_LOAD_LEVEL		100
#define	DEFAULT_VH_LEVEL	 80
#define	DEFAULT_H_LEVEL		 66
#define	DEFAULT_L_LEVEL		 33
#define	DEFAULT_VL_LEVEL	 15
#define	MIN_LOAD_LEVEL		  0

#define	MIN_SPEED		  1

#define	DEFAULT_LARGE_STEP	  2
#define	MIN_LARGE_STEP		  1

#define	NORMAL_STEP		  1


static int oid_cp_time[CTL_MAXNAME];
static size_t oid_cp_time_len = CTL_MAXNAME;
static int oid_cpu_speed[CTL_MAXNAME];
static size_t oid_cpu_speed_len = CTL_MAXNAME;
static int oid_economy_speed[CTL_MAXNAME];
static size_t oid_economy_speed_len = CTL_MAXNAME;
static int oid_acline[CTL_MAXNAME];
static size_t oid_acline_len = CTL_MAXNAME;
static long cpu_speed[1];
static int len = sizeof(cpu_speed);
static long load[MAX_NUM_OF_DATA], sum;
static int curr_pos, data_cnt;
static int num_of_data, interval, verbose_f;
static int very_low_level, low_level, high_level, very_high_level, large_step;
static int max_speed, cpu_max_speed;


#define	get_sysctl_param(__var__) \
		((sysctl(oid_ ## __var__, oid_ ## __var__ ## _len, __var__, &len, NULL, 0) < 0) \
		 ? (err(3, "get sysctl: %s", #__var__), -1) \
		 : (int)__var__[0])

#define	set_sysctl_param(__var__, __value__) \
		if (sysctl(oid_ ## __var__, oid_ ## __var__ ## _len, NULL, 0, __value__, len) < 0) { \
			err(4, "set sysctl: %s", #__var__); \
		}


#define	DPRINTF(__args__)	{ if (verbose_f) printf __args__; }

int main(int argc, char *argv[])
{
    long avg;
    int c, error_f;
    int online;
    int curr_speed, prev_speed, new_speed, step;

    void init_mib(void);
    void init_loads(void);
    int get_acline_status(void);
    int get_cpu_speed(void);
    long get_cpu_load(void);
    void set_cpu_speed(int speed);
    int param_check(int value, int min, int max, int default_value,
		    const char *name);
    void usage(void);
    void show_params(void);


    /* initialize sysctl MIBs */
    init_mib();

    /* setup parameters to default */
    interval = DEFAULT_INTERVAL;
    num_of_data = DEFAULT_NUM_OF_DATA;
    very_high_level = DEFAULT_VH_LEVEL;
    high_level = DEFAULT_H_LEVEL;
    low_level = DEFAULT_L_LEVEL;
    very_low_level = DEFAULT_VL_LEVEL;
    large_step = DEFAULT_LARGE_STEP;
    max_speed = cpu_max_speed;
    verbose_f = 0;
    error_f = 0;

    /*
     * perse command line options
     */
    while ((c = getopt(argc, argv, "h:H:i:l:L:m:n:s:uv?")) != EOF) {
	switch (c) {
	  case 'h':
	    high_level = param_check(atoi(optarg),
				     MIN_LOAD_LEVEL, MAX_LOAD_LEVEL,
				     high_level,
				     "high level value");
	    break;
	  case 'H':
	    very_high_level = param_check(atoi(optarg),
					  MIN_LOAD_LEVEL, MAX_LOAD_LEVEL,
					  very_high_level,
					  "very high level value");
	    break;
	  case 'i':
	    interval = param_check(atoi(optarg),
				   MIN_INTERVAL, INT_MAX,
				   interval,
				   "interval value");
	    break;
	  case 'l':
	    low_level = param_check(atoi(optarg),
				    MIN_LOAD_LEVEL, MAX_LOAD_LEVEL,
				    low_level,
				    "low level value");
	    break;
	  case 'L':
	    very_low_level = param_check(atoi(optarg),
					 MIN_LOAD_LEVEL, MAX_LOAD_LEVEL,
					 very_low_level,
					 "very low level value");
	    break;
	  case 'm':
	    max_speed = param_check(atoi(optarg),
				    MIN_SPEED, cpu_max_speed,
				    max_speed,
				    "maximum speed");
	    break;
	  case 'n':
	    num_of_data = param_check(atoi(optarg),
				      MIN_NUM_OF_DATA, MAX_NUM_OF_DATA,
				      num_of_data,
				      "number of data to calculate average");
	    break;
	  case 's':
	    large_step = param_check(atoi(optarg),
				     MIN_LARGE_STEP, INT_MAX,
				     large_step,
				     "large step value");
	    break;
	  case 'u':
	  case '?':
	    ++error_f;
	    break;
	  case 'v':
	    ++verbose_f;
	    break;
	  default:
	    warnx("unknown option `%d'; ignored", c);
	    break;
	}
    }

    /* check water level values */
    if (!(very_low_level < low_level
	  && low_level < high_level && high_level < very_high_level)) {
	warnx("threshold levels must be the following relation: "
	      "%d <= very_low(%d) < low(%d)"
	      " < high(%d) < very_high(%d) <= %d"
	      "; abort.",
	      MIN_LOAD_LEVEL, very_low_level, low_level,
	      high_level, very_high_level, MAX_LOAD_LEVEL);
	++error_f;
    }

    /* check step range */
    if (max_speed <= large_step) {
	warnx("large step value(%d) must be"
	      " between %d to max speed(%d) - 1; abort.",
	      large_step, MIN_LARGE_STEP, max_speed);
	++error_f;
    }
	

    if (error_f) {
	usage();
	exit(-1);
    }

    if (verbose_f) {
	show_params();
    } else {
	if (daemon(0, 1)) {
	    errx(2, "fork failed");
	}
    }

    /* initialize load value */
    get_cpu_load();


    /*
     * main loop
     */
    prev_speed = -1;
    while (1) {
	sleep(interval);

	/* get current status */
	if ((curr_speed = get_cpu_speed()) != prev_speed) {
	    init_loads();
	}
	online = get_acline_status();

	/* if CPU speed is too fast, limit to max_speed */
	if (max_speed < curr_speed && !online) {
	    if (verbose_f) {
		warnx("CPU speed is too fast; set to %d\n", max_speed);
	    }
	    set_cpu_speed(max_speed);
	    continue;	/* no more process in this cycle */
	}

	/* save current status */
	prev_speed = curr_speed;

	/* calculate moving load average */
	sum -= load[curr_pos];
	load[curr_pos] = get_cpu_load();
	sum += load[curr_pos];
	if (data_cnt < num_of_data) {
	    ++data_cnt;
	}
	avg = sum / data_cnt;

	/* increment data pointer */
	if (num_of_data <= ++curr_pos) {
	    curr_pos = 0;
	}

	/* process for load average */
	DPRINTF(("CPU load = %ld (avg by %d)\n", avg, data_cnt));
	if (!online && data_cnt == num_of_data) {
	    DPRINTF(("current CPU speed = %d, "
		     "min speed = %d, max speed = %d\n",
		     curr_speed, MIN_SPEED, max_speed));
	    step = NORMAL_STEP;
	    if (MIN_SPEED < curr_speed) {
		if (avg < low_level) {
		    /* set decrement step */
		    if (avg < very_low_level) {
			step = large_step;
			DPRINTF(("VERY "));
		    }
		    DPRINTF(("LOW\n"));
		    /* speed down */
		    new_speed = curr_speed - step;
		    if (new_speed < MIN_SPEED) {
			new_speed = MIN_SPEED;
		    }
		    set_cpu_speed(new_speed);
		    continue;
		}
	    }
	    if (curr_speed < max_speed) {
		if (high_level < avg) {
		    /* set increment step */
		    if (very_high_level < avg) {
			step = large_step;
			DPRINTF(("VERY "));
		    }
		    DPRINTF(("HIGH\n"));
		    /* speed up */
		    new_speed = curr_speed + step;
		    if (max_speed < new_speed) {
			new_speed = max_speed;
		    }
		    set_cpu_speed(new_speed);
		}
	    }
	}
    }
}


#define	get_mib(__name__, __var__) \
		if (sysctlnametomib(__name__, oid_ ## __var__, &oid_ ## __var__ ## _len) < 0) { \
			err(1, "sysctlname2mib: %s", __name__); \
		}

void init_mib(void)
{
    get_mib("kern.cp_time", cp_time);
    get_mib("hw.acpi.cpu.max_speed", cpu_speed);
    cpu_max_speed = get_sysctl_param(cpu_speed);
    get_mib("hw.acpi.cpu.current_speed", cpu_speed);
    get_mib("hw.acpi.cpu.economy_speed", economy_speed);
    get_mib("hw.acpi.acline", acline);
}


void init_loads(void)
{
    int cnt;

    for (cnt = 0; cnt < num_of_data; cnt++) {
	load[cnt] = 0;
    }
    curr_pos = 0;
    data_cnt = 0;
    sum = 0;
}


int get_acline_status(void)
{
    long acline[1];
    int len = sizeof(acline);

    return get_sysctl_param(acline);
}


int get_cpu_speed(void)
{
    return get_sysctl_param(cpu_speed);
}


long get_cpu_load(void)
{
    long cp_time[CPUSTATES], idle_diff, total, total_diff, load;
    int len = sizeof(cp_time);
    static long old_total = 0, old_idle = 0;

    get_sysctl_param(cp_time);
    idle_diff = cp_time[CP_IDLE] - old_idle;
    total = cp_time[CP_USER] + cp_time[CP_NICE] + cp_time[CP_SYS]
	+ cp_time[CP_INTR] + cp_time[CP_IDLE];
    total_diff = total - old_total;
    load = (total_diff - idle_diff) * MAX_LOAD_LEVEL / total_diff;
    if (verbose_f) {
	if (old_total == 0) {
	    /* first time */
	    printf("CPU load values are initialized"
		   " (total = %ld, idle = %ld)\n", total, cp_time[CP_IDLE]);
	} else {
	    printf("total = %ld, idle = %ld, usage = %ld%%\n",
		   total_diff, idle_diff, load);
	}
    }
    old_total = total;
    old_idle = cp_time[CP_IDLE];

    return load;
}


void set_cpu_speed(int speed)
{
    cpu_speed[0] = speed;
    set_sysctl_param(economy_speed, cpu_speed);
    DPRINTF(("set CPU speed to %d\n", speed));
}


int param_check(int value, int min, int max, int default_value,
		const char *name)
{
    int ret = value;

    if (value < min) {
	warnx("%s (%d) must be `=> %d'; ignored", name, value, min);
	ret = default_value;
    } else if (max < value) {
	warnx("%s (%d) must be `<= %d'; ignored", name, value, max);
	ret = default_value;
    }

    return ret;
}


void usage(void)
{
    fprintf(stderr, "Usage: %s [options]\n", getprogname());
    fprintf(stderr,
	    "    options: -i <value>: interval time for get cpu info. (default = %d)\n",
	    DEFAULT_INTERVAL);
    fprintf(stderr,
	    "             -n <value>: number of load data for calculate moving average (default = %d)\n",
	    DEFAULT_NUM_OF_DATA);
    fprintf(stderr,
	    "             -H <value>: very high level of load (default = %d)\n",
	    DEFAULT_VH_LEVEL);
    fprintf(stderr,
	    "             -h <value>: high level of load (default = %d)\n",
	    DEFAULT_H_LEVEL);
    fprintf(stderr,
	    "             -l <value>: low level of load (default = %d)\n",
	    DEFAULT_L_LEVEL);
    fprintf(stderr,
	    "             -L <value>: very low level of load (default = %d)\n",
	    DEFAULT_VL_LEVEL);
    fprintf(stderr,
	    "             -s <value>: speed inc/dec step CPU load is very high/low (default = %d)\n",
	    DEFAULT_LARGE_STEP);
    fprintf(stderr,
	    "             -m <value>: max speed for economy mode (default = %d)\n",
	    cpu_max_speed);
    fprintf(stderr,
	    "             -v:         verbose mode, don't change to daemon process\n");
    fprintf(stderr,
	    "             -u:         show this message\n");
    fprintf(stderr,
	    "             -?:         show this message\n");
}


void show_params(void)
{
    fprintf(stderr,
	    "Parameters: interval                            = %d\n",
	    interval);
    fprintf(stderr,
	    "            # of data for calculate avarage     = %d\n",
	    num_of_data);
    fprintf(stderr,
	    "            very high load level                = %d\n",
	    very_high_level);
    fprintf(stderr,
	    "            high load level                     = %d\n",
	    high_level);
    fprintf(stderr,
	    "            low load level                      = %d\n",
	    low_level);
    fprintf(stderr,
	    "            very low load level                 = %d\n",
	    very_low_level);
    fprintf(stderr,
	    "            inc/dec step for very high/low load = %d\n",
	    large_step);
    fprintf(stderr,
	    "            max speed for economy mode          = %d\n",
	    max_speed);
}
